from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed.dbapi import DBAPIConnection

from pathlib import Path


def init_schema(connection: DBAPIConnection) -> None:
    schema_script = Path(__file__).parent.parent.parent.parent / "init_schema.sql"

    with open(schema_script, "r") as f:
        script = f.read()

    cursor = connection.cursor()
    cursor.execute(script)
    cursor.close()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed.dbapi import DBAPIConnection

from expat.exceptions import MigrationValidationError
from expat.utils import hash_file

from pathlib import Path
from pydantic import BaseModel
from typing import Self, Type


class Migration(BaseModel):
    migration_path: Path
    migration_id: str
    migration_name: str
    up_hash: str
    down_hash: str

    @classmethod
    def from_path(cls: Type[Self], path: Path) -> Self:
        migration_id, migration_name = path.name.split("_", 1)

        migration = cls(
            migration_path=path,
            migration_id=migration_id,
            migration_name=migration_name,
            up_hash=hash_file(path / "up.sql"),
            down_hash=hash_file(path / "down.sql"),
        )

        return migration

    def apply(
        self,
        connection: DBAPIConnection,
        metadata_connection: DBAPIConnection,
        validate_hashes: bool,
    ) -> None:
        if self._is_applied(metadata_connection, validate_hashes):
            return None

        self._execute_up(connection)
        self._insert_migration_row(metadata_connection)

    def rollback(
        self,
        connection: DBAPIConnection,
        metadata_connection: DBAPIConnection,
        validate_hashes: bool,
    ) -> None:
        if not self._is_applied(metadata_connection, validate_hashes):
            return None

        self._execute_down(connection)
        self._delete_migration_row(metadata_connection)

    def _is_applied(self, connection: DBAPIConnection, validate_hashes: bool) -> bool:
        cursor = connection.cursor()

        cursor.execute(
            f"SELECT * FROM migrations WHERE migration_id = '{self.migration_id}';"
        )
        result = cursor.fetchone()

        cursor.close()

        if result is None:
            return False

        if not isinstance(result, tuple):
            raise ValueError(
                f"Database connections should be configured to return tuples, not {type(result)}"
            )

        migration_row = MigrationRow.from_row(result)

        if validate_hashes:
            if (
                self.up_hash != migration_row.up_hash
                or self.down_hash != migration_row.down_hash
            ):
                raise MigrationValidationError(self, migration_row)

        return True

    def _insert_migration_row(self, connection: DBAPIConnection) -> None:
        cursor = connection.cursor()

        # Have to use f-strings because parameter styles vary by library; not too worried about
        # security here because migrations are usually executed with admin permissions anyway, so
        # no risk of unwanted SQL injection.
        cursor.execute(
            f"""
                INSERT INTO migrations VALUES (
                    '{self.migration_id}',
                    '{self.migration_name}',
                    '{self.up_hash}',
                    '{self.down_hash}'
                );
            """
        )

        cursor.close()

    def _delete_migration_row(self, connection: DBAPIConnection) -> None:
        cursor = connection.cursor()

        cursor.execute(
            f"""
                DELETE FROM migrations
                WHERE migration_id = '{self.migration_id}';
            """
        )

        cursor.close()

    def _execute_up(self, connection: DBAPIConnection) -> None:
        cursor = connection.cursor()

        with open(self.migration_path / "up.sql", "r") as f:
            up_script = f.read()

        cursor.execute(up_script)

        cursor.close()

    def _execute_down(self, connection: DBAPIConnection) -> None:
        cursor = connection.cursor()

        with open(self.migration_path / "down.sql", "r") as f:
            down_script = f.read()

        cursor.execute(down_script)

        cursor.close()


class MigrationRow(BaseModel):
    migration_id: str
    migration_name: str
    up_hash: str
    down_hash: str

    @classmethod
    def from_row(cls: Type[Self], row: tuple[str, str, str, str]) -> Self:
        return cls(
            migration_id=row[0],
            migration_name=row[1],
            up_hash=row[2],
            down_hash=row[3],
        )


def get_migrations(migration_dir: Path) -> list[Migration]:
    migrations: list[Migration] = []

    for dir in sorted(migration_dir.iterdir()):
        if not dir.is_dir():
            continue

        migration = Migration.from_path(dir)

        migrations.append(migration)

    return migrations



from expat.config import ExpatConfig
from expat.db.migration import get_migrations
from expat.db.migration import Migration, MigrationRow

import sqlite3


def test_get_migrations(test_config: ExpatConfig):
    migrations = get_migrations(test_config.migration_dir)

    assert migrations == [
        Migration(
            migration_path=test_config.migration_dir
            / "20250901113604839_test-migration-1",
            migration_id="20250901113604839",
            migration_name="test-migration-1",
            up_hash="81d645ffc570f1fbf7ae82c9faf6edf7",
            down_hash="b8dc9950943caa3acc3578b2defe60d2",
        ),
        Migration(
            migration_path=test_config.migration_dir
            / "20250901114236057_test-migration-2",
            migration_id="20250901114236057",
            migration_name="test-migration-2",
            up_hash="d41d8cd98f00b204e9800998ecf8427e",
            down_hash="d41d8cd98f00b204e9800998ecf8427e",
        ),
    ]


def test_migration_insert(
    test_migrations: list[Migration], initialized_sqlite_db: sqlite3.Connection
):
    migration = test_migrations[0]

    migration._insert_migration_row(initialized_sqlite_db)

    cursor = initialized_sqlite_db.cursor()
    cursor.execute("SELECT * FROM migrations;")
    result = cursor.fetchall()
    cursor.close()

    migrations = [MigrationRow.from_row(row) for row in result]

    assert migrations == [
        MigrationRow(
            migration_id="20250901113604839",
            migration_name="test-migration-1",
            up_hash="81d645ffc570f1fbf7ae82c9faf6edf7",
            down_hash="b8dc9950943caa3acc3578b2defe60d2",
        )
    ]


def test_migration_delete(
    test_migrations: list[Migration], initialized_sqlite_db: sqlite3.Connection
):
    migration = test_migrations[0]
    migration._insert_migration_row(initialized_sqlite_db)
    migration._delete_migration_row(initialized_sqlite_db)

    cursor = initialized_sqlite_db.cursor()
    cursor.execute("SELECT * FROM migrations;")
    result = cursor.fetchall()
    cursor.close()

    assert len(result) == 0


def test_migration_execute_up(
    test_migrations: list[Migration], initialized_sqlite_db: sqlite3.Connection
):
    migration = test_migrations[0]
    migration._execute_up(initialized_sqlite_db)

    cursor = initialized_sqlite_db.cursor()
    cursor.execute("""
        SELECT
            name,
            type,
            pk
        FROM pragma_table_info('up_table');
    """)
    result = cursor.fetchall()
    cursor.close()

    assert result == [
        ("val", "INT", 0),
    ]


def test_migration_execute_down(
    test_migrations: list[Migration], initialized_sqlite_db: sqlite3.Connection
):
    migration = test_migrations[0]
    migration._execute_down(initialized_sqlite_db)

    cursor = initialized_sqlite_db.cursor()
    cursor.execute("""
        SELECT
            name,
            type,
            pk
        FROM pragma_table_info('down_table');
    """)
    result = cursor.fetchall()
    cursor.close()

    assert result == [
        ("val", "INT", 0),
    ]


def test_migration_is_applied(
    test_migrations: list[Migration], initialized_sqlite_db: sqlite3.Connection
):
    migration = test_migrations[0]

    migration._insert_migration_row(initialized_sqlite_db)

    is_applied = migration._is_applied(initialized_sqlite_db, True)

    assert is_applied



import hashlib
from pathlib import Path


def hash_file(path: Path) -> str:
    hasher = hashlib.md5()

    with path.open("rb") as f:
        file_content = f.read()

    hasher.update(file_content)

    return hasher.hexdigest()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed.dbapi import DBAPIConnection

from expat.config import ExpatConfig
from expat.db.init_schema import init_schema as _init_schema
from expat.db.migration import get_migrations as _get_migrations


def up(
    connection: DBAPIConnection,
    config: ExpatConfig,
    metadata_connection: DBAPIConnection | None = None,
) -> None:
    """Bring database up to date with current state of migration directory.

    In other words, execute all `up.sql` files in migration directory sequentially from smallest ID
    to largest ID.

    Args:
        connection (DBAPIConnection): A database connection object compliant with PEP 249's DB API
            specification.
        config (ExpatConfig): Expat configuration object.
        metadata_connection (DBAPIConnection | None, optional): Optional additional connection used
            only for managing metadata. Defaults to None.
    """
    migrations = _get_migrations(config.migration_dir)

    if metadata_connection is None:
        metadata_connection = connection

    _init_schema(metadata_connection)

    for migration in migrations:
        migration.apply(connection, metadata_connection, config.validate_hashes)


def down(
    connection: DBAPIConnection,
    config: ExpatConfig,
    metadata_connection: DBAPIConnection | None = None,
) -> None:
    """Tear down database.

    In other words, execute all `down.sql` files in migration directory sequentiallyu from largest
    ID to smallest ID.

    Args:
        connection (DBAPIConnection): A database connection object compliant with PEP 249's DB API
            specification.
        config (ExpatConfig): Expat configuration object.
        metadata_connection (DBAPIConnection | None, optional): Optional additional connection used
            only for managing metadata.
    """
    migrations = _get_migrations(config.migration_dir)
    migrations.reverse()

    if metadata_connection is None:
        metadata_connection = connection

    for migration in migrations:
        migration.rollback(connection, metadata_connection, config.validate_hashes)



from pathlib import Path
from pydantic import BaseModel, Field


class ExpatConfig(BaseModel):
    """Primary configuration class for expat.

    Attributes:
        migration_dir (Path, optional): Directory containing migration files. Defaults to
            `./migrations`.
        validate_hashes (bool, optional): Whether or not to verify that migration files in the
            migration directory match the migrations which have been applied to the database. If
            True, expat will raise an error when it encounters a migration which has been applied
            to the database but has since been tampered with. Defaults to True.
    """

    migration_dir: Path = Field(default=Path() / "migrations")
    validate_hashes: bool = Field(default=True)



from expat.config import ExpatConfig
from expat.db.init_schema import init_schema
from expat.db.migration import Migration, get_migrations

from pathlib import Path
from pytest import fixture
import sqlite3
from typing import Any, Generator


@fixture
def test_config() -> ExpatConfig:
    return ExpatConfig(
        migration_dir=Path(__file__).parent.parent.parent / "test_data" / "migrations"
    )


@fixture
def sqlite_db() -> Generator[sqlite3.Connection, Any, Any]:
    db_path = Path("test.db")
    connection = sqlite3.connect(db_path)

    yield connection

    connection.close()
    db_path.unlink()


@fixture
def initialized_sqlite_db() -> Generator[sqlite3.Connection, Any, Any]:
    db_path = Path("test.db")
    connection = sqlite3.connect(db_path)
    init_schema(connection)

    yield connection

    connection.close()
    db_path.unlink()


@fixture
def test_migrations(test_config: ExpatConfig) -> list[Migration]:
    return get_migrations(test_config.migration_dir)



from expat.db import init_schema

import sqlite3


def test_schema_up(sqlite_db: sqlite3.Connection):
    init_schema.init_schema(sqlite_db)

    cursor = sqlite_db.cursor()
    cursor.execute("""
        SELECT
            name,
            type,
            pk
        FROM pragma_table_info('migrations');
    """)
    result = cursor.fetchall()
    cursor.close()

    assert result == [
        ("migration_id", "VARCHAR(17)", 1),
        ("migration_name", "VARCHAR(1000)", 0),
        ("up_hash", "VARCHAR(32)", 0),
        ("down_hash", "VARCHAR(32)", 0),
    ]



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from expat.db.migration import Migration, MigrationRow


class MigrationValidationError(Exception):
    def __init__(self, repo_migration: Migration, db_migration: MigrationRow):
        self.message = (
            "Migration in repo does not match that which has been applied to database:\n"
            f"    repo: (up_hash: {repo_migration.up_hash}, down_hash: {repo_migration.down_hash})"
            f"\n      db: (up_hash: {db_migration.up_hash}, down_hash: {db_migration.down_hash})"
        )
